import logging
from dataclasses import dataclass
from datetime import UTC, datetime
from threading import Event, Lock, Thread

from .weather import Forecast, WeatherFetchError, fetch_forecast_data


@dataclass(frozen=True, slots=True)
class ForecastLoading:
    message: str = "Loading weather data"


@dataclass(frozen=True, slots=True)
class ForecastReady:
    forecast: Forecast
    updated_at: datetime


@dataclass(frozen=True, slots=True)
class ForecastFailedNoData:
    message: str


@dataclass(frozen=True, slots=True)
class ForecastFailedStale:
    message: str
    last_successful_update: datetime


type ForecastState = (
    ForecastLoading | ForecastReady | ForecastFailedNoData | ForecastFailedStale
)


_forecast_lock = Lock()
_latest_forecast_state: list[ForecastState] = [ForecastLoading()]
logger = logging.getLogger(__name__)


def get_forecast_state() -> ForecastState:
    with _forecast_lock:
        return _latest_forecast_state[0]


def clear_latest_forecast() -> None:
    with _forecast_lock:
        _latest_forecast_state[0] = ForecastLoading()


def _last_successful_update(state: ForecastState) -> datetime | None:
    if isinstance(state, ForecastReady):
        return state.updated_at
    if isinstance(state, ForecastFailedStale):
        return state.last_successful_update
    return None


def update_forecast(latitude: float, longitude: float) -> None:
    try:
        forecast = fetch_forecast_data(latitude, longitude)
    except WeatherFetchError as exc:
        with _forecast_lock:
            previous_state = _latest_forecast_state[0]
            last_successful_update = _last_successful_update(previous_state)
            if last_successful_update is not None:
                _latest_forecast_state[0] = ForecastFailedStale(
                    message=str(exc),
                    last_successful_update=last_successful_update,
                )
            else:
                _latest_forecast_state[0] = ForecastFailedNoData(message=str(exc))
        logger.warning(
            "Forecast update failed (latitude=%s, longitude=%s): %s",
            latitude,
            longitude,
            exc,
        )
        return

    updated_at = datetime.now(tz=UTC)
    with _forecast_lock:
        _latest_forecast_state[0] = ForecastReady(
            forecast=forecast, updated_at=updated_at
        )

    logger.info(
        "Forecast updated (latitude=%s, longitude=%s, points=%s)",
        latitude,
        longitude,
        len(forecast.time),
    )


def run_forecast_updater(
    latitude: float,
    longitude: float,
    weather_update_secs: int,
    stop_event: Event,
) -> None:
    logger.info("Forecast updater loop started")
    while not stop_event.is_set():
        update_forecast(latitude, longitude)
        if stop_event.wait(weather_update_secs):
            break
    logger.info("Forecast updater loop stopped")


def start_forecast_updater(
    latitude: float,
    longitude: float,
    weather_update_secs: int,
) -> tuple[Thread, Event]:
    logger.info(
        "Starting forecast updater (interval=%ss, latitude=%s, longitude=%s)",
        weather_update_secs,
        latitude,
        longitude,
    )
    stop_event = Event()
    updater_thread = Thread(
        target=run_forecast_updater,
        args=(latitude, longitude, weather_update_secs, stop_event),
        daemon=True,
    )
    updater_thread.start()
    return updater_thread, stop_event


def stop_forecast_updater(stop_event: Event, _updater_thread: Thread) -> None:
    stop_event.set()
    # We deliberately do not care about joining the daemon updater thread.
    logger.info("Signalled forecast updater to stop")



from datetime import timedelta
from typing import TYPE_CHECKING, cast

from .chart_context import CHART_HOURS
from .common import COLOUR_BLACK, COLOUR_WHITE, ChartBounds, build_vertical_band

if TYPE_CHECKING:
    from datetime import datetime

    from PIL import ImageDraw

    from kindle_dashboard.weather import Forecast

    from .chart_context import ChartRenderContext

CLOUD_COVER_PLOT_HEIGHT_RATIO = 0.05


def clamp_percentage(value: float) -> float:
    return max(0.0, min(100.0, value))


def combine_cloud_cover(low_cloud_cover: float, mid_cloud_cover: float) -> float:
    return clamp_percentage(max(low_cloud_cover, mid_cloud_cover / 2.0))


def cloud_cover_to_greyscale(cloud_cover: float) -> int:
    clamped_cloud_cover = clamp_percentage(cloud_cover)
    return round(
        COLOUR_WHITE - ((clamped_cloud_cover / 100.0) * (COLOUR_WHITE - COLOUR_BLACK))
    )


def _build_strict_hourly_series(
    forecast: Forecast,
    values: list[float],
    hour_slots: list[datetime],
    *,
    hour_shift: int,
    series_name: str,
) -> list[float]:
    if len(hour_slots) != CHART_HOURS:
        msg = "Cloud cover renderer requires a 24-hour display window"
        raise ValueError(msg)

    lookup: dict[datetime, float] = {}
    shift_delta = timedelta(hours=hour_shift)
    for timestamp, value in zip(forecast.hourly_datetimes, values, strict=False):
        shifted_timestamp = timestamp + shift_delta
        lookup[shifted_timestamp.replace(minute=0, second=0, microsecond=0)] = value

    missing = object()
    series: list[float] = []
    missing_slots: list[datetime] = []
    for slot in hour_slots:
        value = lookup.get(slot, missing)
        if value is missing:
            missing_slots.append(slot)
            continue
        series.append(cast("float", value))

    if missing_slots:
        first_missing_slot = missing_slots[0].isoformat(timespec="minutes")
        msg = (
            f"Cloud cover series '{series_name}' did not provide {CHART_HOURS} points "
            "for the display window; "
            f"missing {len(missing_slots)} point(s) "
            f"(first missing slot: {first_missing_slot})"
        )
        raise ValueError(msg)

    return series


def build_cloud_cover_series(
    forecast: Forecast,
    hour_slots: list[datetime],
) -> list[float]:
    # Open-Meteo documents low/mid cloud cover as instantaneous hourly values,
    # so we draw each timestamp's value from that hour onwards.
    slot_low_cloud_cover = _build_strict_hourly_series(
        forecast,
        forecast.cloudcover_low,
        hour_slots,
        hour_shift=0,
        series_name="cloudcover_low",
    )
    slot_mid_cloud_cover = _build_strict_hourly_series(
        forecast,
        forecast.cloudcover_mid,
        hour_slots,
        hour_shift=0,
        series_name="cloudcover_mid",
    )
    return [
        combine_cloud_cover(low_cloud_cover, mid_cloud_cover)
        for low_cloud_cover, mid_cloud_cover in zip(
            slot_low_cloud_cover, slot_mid_cloud_cover, strict=False
        )
    ]


def draw_cloud_cover_chart(
    draw: ImageDraw.ImageDraw,
    chart_bounds: ChartBounds,
    chart_context: ChartRenderContext,
    forecast: Forecast,
) -> None:
    slot_cloud_cover = build_cloud_cover_series(forecast, chart_context.hour_slots)
    cloud_cover_top, cloud_cover_bottom = build_vertical_band(
        chart_bounds.top,
        chart_bounds.bottom,
        top_ratio=0.0,
        bottom_ratio=CLOUD_COVER_PLOT_HEIGHT_RATIO,
    )

    for index, cloud_cover in enumerate(slot_cloud_cover[:-1]):
        interval_left = chart_context.x_positions[index]
        interval_right = chart_context.x_positions[index + 1]
        fill_colour = cloud_cover_to_greyscale(cloud_cover)
        draw.rectangle(
            (interval_left, cloud_cover_top, interval_right, cloud_cover_bottom),
            fill=fill_colour,
            outline=fill_colour,
        )



import io
import logging
import math
from datetime import UTC, datetime
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from typing import TYPE_CHECKING
from urllib.parse import parse_qs, urlparse

from .config import load_server_config
from .forecast_updater import (
    ForecastFailedNoData,
    ForecastFailedStale,
    ForecastLoading,
    ForecastReady,
    clear_latest_forecast,
    get_forecast_state,
    start_forecast_updater,
    stop_forecast_updater,
)
from .logging_config import configure_logging
from .render import render_dashboard

if TYPE_CHECKING:
    from .weather import Forecast

logger = logging.getLogger(__name__)


def _render_dashboard_png_payload(
    width: int,
    height: int,
    battery_level: int | None,
    supersample_scale: int,
    forecast: Forecast | None,
    forecast_error_message: str | None,
    *,
    weather_configured: bool,
    weather_data_stale: bool,
    weather_data_stale_hours: int | None,
    next_rain_gap_tolerance_hours: float,
) -> bytes:
    image = render_dashboard(
        width,
        height,
        battery_level=battery_level,
        forecast=forecast,
        forecast_error_message=forecast_error_message,
        weather_configured=weather_configured,
        weather_data_stale=weather_data_stale,
        weather_data_stale_hours=weather_data_stale_hours,
        supersample_scale=supersample_scale,
        next_rain_gap_tolerance_hours=next_rain_gap_tolerance_hours,
    )
    # Render final image as greyscale PNG.
    buffer = io.BytesIO()
    image.save(buffer, format="PNG", optimize=False, compress_level=6)
    return buffer.getvalue()


def _normalise_forecast_state(
    forecast_state: ForecastLoading
    | ForecastReady
    | ForecastFailedNoData
    | ForecastFailedStale,
) -> tuple[Forecast | None, str | None, datetime | None]:
    if isinstance(forecast_state, ForecastReady):
        return forecast_state.forecast, None, forecast_state.updated_at
    if isinstance(forecast_state, ForecastFailedStale):
        return None, forecast_state.message, forecast_state.last_successful_update
    if isinstance(forecast_state, ForecastFailedNoData):
        return None, forecast_state.message, None
    return None, None, None


class DashboardHandler(BaseHTTPRequestHandler):
    default_width: int
    default_height: int
    render_supersample_scale: int
    weather_update_secs: int
    next_rain_gap_tolerance_hours: float
    weather_configured: bool
    _request_battery_level: int | None = None
    _response_png_bytes: int | None = None

    def do_GET(self) -> None:
        parsed = urlparse(self.path)
        if parsed.path != "/":
            logger.warning("Received request for unknown path: %s", parsed.path)
            self.send_error(HTTPStatus.NOT_FOUND, "Not found")
            return

        self._serve_dashboard_png(parsed.query)

    def _serve_dashboard_png(self, raw_query: str) -> None:
        raw_battery_level = parse_qs(raw_query).get("batteryLevel", [None])[-1]
        battery_level = int(raw_battery_level) if raw_battery_level else None
        (
            forecast,
            forecast_error_message,
            forecast_updated_at,
        ) = _normalise_forecast_state(get_forecast_state())
        weather_data_stale = False
        weather_data_stale_hours: int | None = None
        if self.weather_configured and forecast_updated_at is not None:
            weather_data_age_hours = _weather_data_age_hours(forecast_updated_at)
            weather_data_stale = _is_weather_data_stale(
                forecast_updated_at,
                self.weather_update_secs,
            )
            if weather_data_stale:
                weather_data_stale_hours = max(1, math.ceil(weather_data_age_hours))
        payload = _render_dashboard_png_payload(
            self.default_width,
            self.default_height,
            battery_level=battery_level,
            supersample_scale=self.render_supersample_scale,
            forecast=forecast,
            forecast_error_message=forecast_error_message,
            weather_configured=self.weather_configured,
            weather_data_stale=weather_data_stale,
            weather_data_stale_hours=weather_data_stale_hours,
            next_rain_gap_tolerance_hours=self.next_rain_gap_tolerance_hours,
        )
        self._request_battery_level = battery_level
        payload_size = len(payload)
        self._response_png_bytes = payload_size

        self.send_response(HTTPStatus.OK)
        self.send_header("Content-Type", "image/png")
        self.send_header("Content-Length", str(payload_size))
        self.end_headers()
        self.wfile.write(payload)

    def log_message(self, format: str, *args: object) -> None:  # noqa: A002
        message = format % args
        if self._response_png_bytes is not None:
            logger.info(
                '%s - "%s" (png_bytes=%s, battery_level=%s)',
                self.client_address[0],
                message,
                self._response_png_bytes,
                self._request_battery_level,
            )
            return

        logger.info('%s - "%s"', self.client_address[0], message)


def _is_weather_data_stale(
    forecast_updated_at: datetime,
    weather_update_secs: int,
) -> bool:
    return _weather_data_age_hours(forecast_updated_at) > (
        float(weather_update_secs) / 3600.0
    )


def _weather_data_age_hours(forecast_updated_at: datetime) -> float:
    return max(
        0.0,
        (datetime.now(tz=UTC) - forecast_updated_at).total_seconds() / 3600.0,
    )


def main() -> None:
    configure_logging()
    logger.info("Starting kindle dashboard service")

    config = load_server_config()

    logger.info("Server configuration loaded")
    clear_latest_forecast()
    logger.info("Forecast state cleared")

    if config.weather_latitude is None:
        logger.warning(
            "WEATHER_LATITUDE and WEATHER_LONGITUDE are not set; "
            "weather data will not be displayed."
        )

    server = ThreadingHTTPServer(
        server_address=(config.host, config.port),
        RequestHandlerClass=DashboardHandler,
    )
    server.daemon_threads = True
    DashboardHandler.default_width = config.default_width
    DashboardHandler.default_height = config.default_height
    DashboardHandler.render_supersample_scale = config.render_supersample_scale
    DashboardHandler.weather_update_secs = config.weather_update_secs
    DashboardHandler.next_rain_gap_tolerance_hours = (
        config.next_rain_gap_tolerance_hours
    )

    updater_thread = None
    stop_event = None
    latitude = config.weather_latitude
    longitude = config.weather_longitude
    if latitude is not None:
        assert longitude is not None
        DashboardHandler.weather_configured = True
        updater_thread, stop_event = start_forecast_updater(
            latitude,
            longitude,
            config.weather_update_secs,
        )
    else:
        DashboardHandler.weather_configured = False
        logger.info("Weather updater is disabled")

    logger.info(
        "Serving kindle dashboard on http://%s:%s/ "
        "(default: %sx%s, supersample: %sx, "
        "png mode: indexed 16-level greyscale)",
        config.host,
        config.port,
        config.default_width,
        config.default_height,
        config.render_supersample_scale,
    )
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        logger.info("Received interrupt signal, shutting down")
    finally:
        server.server_close()
        logger.info("HTTP server stopped")
        if stop_event is not None and updater_thread is not None:
            stop_forecast_updater(stop_event, updater_thread)



from math import ceil, floor
from typing import TYPE_CHECKING

from .cloud_cover import CLOUD_COVER_PLOT_HEIGHT_RATIO
from .common import (
    COLOUR_BLACK,
    ChartBounds,
    FontType,
    RenderMetrics,
    build_vertical_band,
    measure_text,
    measure_text_width,
)
from .forecast import build_hourly_series, build_sun_event_markers

if TYPE_CHECKING:
    from collections.abc import Callable
    from datetime import datetime

    from PIL import ImageDraw

    from kindle_dashboard.weather import Forecast

    from .chart_context import ChartRenderContext

COLOUR_LIGHT_GREY = 191
COLOUR_MID_GREY = 128
TEMPERATURE_PLOT_TOP_RATIO = 0.06
TEMPERATURE_PLOT_BOTTOM_RATIO = 0.65
TEMPERATURE_PLOT_DEFAULT_MIN = 5.0
TEMPERATURE_PLOT_DEFAULT_MAX = 25.0
TEMPERATURE_RULE_INTERVAL_DEGREES = 5
TEMPERATURE_RULE_COLOUR = 223


def format_temperature_label(value: float) -> str:
    return f"{round(value):.0f}\N{DEGREE SIGN}"


def draw_temperature_annotation(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    anchor_x: int,
    anchor_y: int,
    text: str,
    *,
    prefer_above: bool,
    temperature_label_offset: int,
    minimum_y_for_above_label: int | None = None,
) -> None:
    text_width, text_height, _ = measure_text(draw, font, text)

    text_x = anchor_x - (text_width // 2)
    image_width = draw.im.size[0]
    max_text_x = max(0, image_width - text_width)
    text_x = max(0, min(text_x, max_text_x))

    if prefer_above:
        preferred_above_y = anchor_y - text_height - temperature_label_offset
        if (
            minimum_y_for_above_label is not None
            and preferred_above_y < minimum_y_for_above_label
        ):
            text_y = anchor_y + temperature_label_offset
        else:
            text_y = preferred_above_y
    else:
        text_y = anchor_y + temperature_label_offset
    draw.text((text_x, text_y), text, fill=COLOUR_BLACK, font=font)


def draw_temperature_extrema(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    point_locations: dict[int, tuple[int, int]],
    extrema_points: list[tuple[int, float, bool]],
    *,
    marker_radius: int,
    temperature_label_offset: int,
    minimum_y_for_above_label: int | None = None,
) -> None:
    seen_marker_indices: set[int] = set()
    for point_index, point_temperature, is_max in extrema_points:
        point_x, point_y = point_locations[point_index]
        if point_index not in seen_marker_indices:
            draw.ellipse(
                (
                    point_x - marker_radius,
                    point_y - marker_radius,
                    point_x + marker_radius,
                    point_y + marker_radius,
                ),
                fill=COLOUR_BLACK,
                outline=COLOUR_BLACK,
            )
            seen_marker_indices.add(point_index)

        draw_temperature_annotation(
            draw=draw,
            font=font,
            anchor_x=point_x,
            anchor_y=point_y,
            text=format_temperature_label(point_temperature),
            prefer_above=is_max,
            temperature_label_offset=temperature_label_offset,
            minimum_y_for_above_label=minimum_y_for_above_label,
        )


def draw_sunrise_sunset_arrows(
    draw: ImageDraw.ImageDraw,
    chart_bounds: ChartBounds,
    hour_slots: list[datetime],
    forecast: Forecast,
    slot_temperatures: list[float],
    temperature_to_y: Callable[[float], int],
    metrics: RenderMetrics,
) -> None:
    window_start = hour_slots[0]
    window_end = hour_slots[-1]
    sun_markers = build_sun_event_markers(forecast, window_start, window_end)
    span_hours = float(len(hour_slots) - 1)
    for marker_time, is_sunrise in sun_markers:
        offset_hours = (marker_time - window_start).total_seconds() / 3600.0
        if offset_hours < 0 or offset_hours > span_hours:
            continue
        marker_x = chart_bounds.left + round(
            ((chart_bounds.right - chart_bounds.left) * offset_hours) / span_hours
        )
        temperature_line_y = estimate_temperature_line_y(
            marker_time=marker_time,
            window_start=window_start,
            slot_temperatures=slot_temperatures,
            temperature_to_y=temperature_to_y,
        )

        marker_tip_y = (
            temperature_line_y - metrics.temperature_label_offset
            if is_sunrise
            else temperature_line_y + metrics.temperature_label_offset
        )
        draw_directional_arrow(
            draw,
            x=marker_x,
            tip_y=marker_tip_y,
            is_up=is_sunrise,
            metrics=metrics,
        )


def draw_directional_arrow(
    draw: ImageDraw.ImageDraw,
    *,
    x: int,
    tip_y: int,
    is_up: bool,
    metrics: RenderMetrics,
) -> None:
    arrow_head_size = max(2, metrics.thin_shape_stroke_width * 2)
    stem_length = max(4, metrics.tick_length * 2)
    line_width = metrics.thin_shape_stroke_width
    marker_colour = COLOUR_MID_GREY

    if is_up:
        stem_top = tip_y + arrow_head_size
        stem_bottom = stem_top + stem_length
        draw.line((x, stem_top, x, stem_bottom), fill=marker_colour, width=line_width)
        draw.line(
            (x, tip_y, x - arrow_head_size, stem_top),
            fill=marker_colour,
            width=line_width,
        )
        draw.line(
            (x, tip_y, x + arrow_head_size, stem_top),
            fill=marker_colour,
            width=line_width,
        )
        return

    stem_bottom = tip_y - arrow_head_size
    stem_top = stem_bottom - stem_length
    draw.line((x, stem_top, x, stem_bottom), fill=marker_colour, width=line_width)
    draw.line(
        (x, tip_y, x - arrow_head_size, stem_bottom),
        fill=marker_colour,
        width=line_width,
    )
    draw.line(
        (x, tip_y, x + arrow_head_size, stem_bottom),
        fill=marker_colour,
        width=line_width,
    )


def estimate_temperature_line_y(
    *,
    marker_time: datetime,
    window_start: datetime,
    slot_temperatures: list[float],
    temperature_to_y: Callable[[float], int],
) -> int:
    marker_offset_hours = max(
        0.0,
        min(
            float(len(slot_temperatures) - 1),
            (marker_time - window_start).total_seconds() / 3600.0,
        ),
    )
    left_index = int(marker_offset_hours)
    right_index = min(left_index + 1, len(slot_temperatures) - 1)
    interpolation = marker_offset_hours - left_index

    left_temperature = slot_temperatures[left_index]
    right_temperature = slot_temperatures[right_index]
    chosen_temperature = left_temperature + (
        (right_temperature - left_temperature) * interpolation
    )
    return temperature_to_y(chosen_temperature)


def draw_temperature_chart(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    chart_bounds: ChartBounds,
    chart_context: ChartRenderContext,
    forecast: Forecast,
    metrics: RenderMetrics,
) -> None:
    hour_slots = chart_context.hour_slots
    x_axis_y = chart_context.x_axis_y
    x_positions = chart_context.x_positions

    slot_temperatures = build_hourly_series(
        forecast,
        forecast.temperature_2m,
        hour_slots,
        hour_shift=0,
    )

    temperature_band_top, temperature_band_bottom = build_vertical_band(
        chart_bounds.top,
        chart_bounds.bottom,
        top_ratio=TEMPERATURE_PLOT_TOP_RATIO,
        bottom_ratio=TEMPERATURE_PLOT_BOTTOM_RATIO,
    )
    _, cloud_band_bottom = build_vertical_band(
        chart_bounds.top,
        chart_bounds.bottom,
        top_ratio=0.0,
        bottom_ratio=CLOUD_COVER_PLOT_HEIGHT_RATIO,
    )
    plot_top = temperature_band_top + metrics.plot_padding
    plot_bottom = temperature_band_bottom - metrics.plot_padding
    draw.line(
        (
            chart_bounds.left,
            x_axis_y,
            chart_bounds.right,
            x_axis_y,
        ),
        fill=COLOUR_BLACK,
        width=metrics.thin_shape_stroke_width,
    )

    hour_label_widths: dict[str, int] = {}
    for index, slot in enumerate(hour_slots):
        x_position = x_positions[index]
        if slot.hour == 0:
            draw.line(
                (
                    x_position,
                    chart_bounds.top,
                    x_position,
                    x_axis_y,
                ),
                fill=COLOUR_LIGHT_GREY,
                width=metrics.thin_shape_stroke_width,
            )

        draw.line(
            (
                x_position,
                x_axis_y,
                x_position,
                x_axis_y - metrics.tick_length,
            ),
            fill=COLOUR_BLACK,
            width=metrics.thin_shape_stroke_width,
        )

        if slot.hour == 0:
            continue

        hour_label = slot.strftime("%H")
        hour_width = hour_label_widths.get(hour_label)
        if hour_width is None:
            hour_width = measure_text_width(draw, font, hour_label)
            hour_label_widths[hour_label] = hour_width
        draw.text(
            (
                x_position - hour_width // 2,
                x_axis_y + metrics.axis_label_gap,
            ),
            hour_label,
            fill=COLOUR_BLACK,
            font=font,
        )

    observed_min_temperature = min(slot_temperatures)
    observed_max_temperature = max(slot_temperatures)
    first_relevant_rule_temperature = int(
        floor(observed_min_temperature / TEMPERATURE_RULE_INTERVAL_DEGREES)
        * TEMPERATURE_RULE_INTERVAL_DEGREES
    )
    last_relevant_rule_temperature = int(
        ceil(observed_max_temperature / TEMPERATURE_RULE_INTERVAL_DEGREES)
        * TEMPERATURE_RULE_INTERVAL_DEGREES
    )
    min_chart_temperature = min(
        TEMPERATURE_PLOT_DEFAULT_MIN,
        float(first_relevant_rule_temperature),
    )
    max_chart_temperature = max(
        TEMPERATURE_PLOT_DEFAULT_MAX,
        float(last_relevant_rule_temperature),
    )
    value_span = max_chart_temperature - min_chart_temperature
    plot_height = plot_bottom - plot_top

    def temperature_to_y(temperature: float) -> int:
        clamped_temperature = max(
            min_chart_temperature,
            min(temperature, max_chart_temperature),
        )
        normalised = (clamped_temperature - min_chart_temperature) / value_span
        return plot_bottom - round(normalised * plot_height)

    for rule_temperature in range(
        first_relevant_rule_temperature,
        last_relevant_rule_temperature + 1,
        TEMPERATURE_RULE_INTERVAL_DEGREES,
    ):
        normalised = (rule_temperature - min_chart_temperature) / value_span
        rule_y = plot_bottom - round(normalised * plot_height)
        if plot_top <= rule_y <= plot_bottom:
            draw.line(
                (
                    chart_bounds.left,
                    rule_y,
                    chart_bounds.right,
                    rule_y,
                ),
                fill=TEMPERATURE_RULE_COLOUR,
                width=max(1, metrics.thin_shape_stroke_width),
            )

    draw_sunrise_sunset_arrows(
        draw=draw,
        chart_bounds=chart_bounds,
        hour_slots=hour_slots,
        forecast=forecast,
        slot_temperatures=slot_temperatures,
        temperature_to_y=temperature_to_y,
        metrics=metrics,
    )

    line_points = [
        (x_position, temperature_to_y(temperature))
        for x_position, temperature in zip(x_positions, slot_temperatures, strict=False)
    ]
    point_locations = dict(enumerate(line_points))

    draw.line(
        line_points,
        fill=COLOUR_BLACK,
        width=metrics.temperature_plot_line_width,
    )

    min_index, min_temperature = min(
        enumerate(slot_temperatures), key=lambda item: item[1]
    )
    max_index, max_temperature = max(
        enumerate(slot_temperatures), key=lambda item: item[1]
    )
    extrema_points = [
        (min_index, min_temperature, False),
        (max_index, max_temperature, True),
    ]
    draw_temperature_extrema(
        draw=draw,
        font=font,
        point_locations=point_locations,
        extrema_points=extrema_points,
        marker_radius=metrics.marker_radius,
        temperature_label_offset=metrics.temperature_label_offset,
        minimum_y_for_above_label=cloud_band_bottom + metrics.plot_padding,
    )



from dataclasses import dataclass

from PIL import Image, ImageDraw, ImageFont

ERROR_TEXT_LINE_SPACING = 2
THIN_SHAPE_FINAL_STROKE_WIDTH = 1.5
KINDLE_GREYSCALE_LEVELS = 16
KINDLE_GREYSCALE_STEP = 255 / (KINDLE_GREYSCALE_LEVELS - 1)

MIN_CHART_DIMENSION = 20
TEMPERATURE_PLOT_LINE_WIDTH = 3
TEMPERATURE_LABEL_OFFSET = 10

TEXT_ELLIPSIS = "..."

COLOUR_BLACK = 0
COLOUR_WHITE = 255

FontType = ImageFont.ImageFont | ImageFont.FreeTypeFont


@dataclass(frozen=True, slots=True)
class RenderMetrics:
    border_width: int
    padding: int
    group_spacing: int
    battery_icon_min_height: int
    battery_icon_text_padding: int
    temperature_plot_line_width: int
    temperature_label_offset: int
    error_text_line_spacing: int
    thin_shape_stroke_width: int
    tick_length: int
    axis_label_gap: int
    plot_padding: int
    marker_radius: int
    min_chart_dimension: int


@dataclass(frozen=True, slots=True)
class ChartBounds:
    left: int
    top: int
    right: int
    bottom: int


def build_render_metrics(width: int, height: int, render_scale: int) -> RenderMetrics:
    scaled_width = width * render_scale
    scaled_height = height * render_scale
    return RenderMetrics(
        border_width=max(1, min(scaled_width, scaled_height) // 200),
        padding=max(8 * render_scale, min(scaled_width, scaled_height) // 60),
        group_spacing=6 * render_scale,
        battery_icon_min_height=12 * render_scale,
        battery_icon_text_padding=2 * render_scale,
        temperature_plot_line_width=TEMPERATURE_PLOT_LINE_WIDTH * render_scale,
        temperature_label_offset=TEMPERATURE_LABEL_OFFSET * render_scale,
        error_text_line_spacing=ERROR_TEXT_LINE_SPACING * render_scale,
        thin_shape_stroke_width=scale_thin_shape_stroke(render_scale),
        tick_length=3 * render_scale,
        axis_label_gap=4 * render_scale,
        plot_padding=2 * render_scale,
        marker_radius=3 * render_scale,
        min_chart_dimension=MIN_CHART_DIMENSION * render_scale,
    )


def build_x_positions(
    chart_left: int,
    chart_right: int,
    point_count: int,
) -> list[int]:
    return [
        chart_left + round((chart_right - chart_left) * index / (point_count - 1))
        for index in range(point_count)
    ]


def build_chart_bounds(
    scaled_width: int,
    scaled_height: int,
    metrics: RenderMetrics,
) -> ChartBounds:
    return ChartBounds(
        left=metrics.border_width + metrics.padding,
        top=metrics.border_width,
        right=scaled_width - metrics.border_width - metrics.padding,
        bottom=scaled_height - metrics.border_width - metrics.padding,
    )


def build_vertical_band(
    chart_top: int,
    chart_bottom: int,
    *,
    top_ratio: float,
    bottom_ratio: float,
) -> tuple[int, int]:
    chart_height = chart_bottom - chart_top + 1
    band_top = chart_top + round((chart_height - 1) * top_ratio)
    band_bottom = chart_top + round((chart_height - 1) * bottom_ratio)
    return band_top, band_bottom


def measure_text(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    text: str,
) -> tuple[int, int, int]:
    left, top, right, bottom = draw.textbbox((0, 0), text, font=font)
    return round(right - left), round(bottom - top), round(top)


def measure_text_width(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    text: str,
) -> int:
    return measure_text(draw, font, text)[0]


def fit_text_to_width(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    text: str,
    max_width: int,
    suffix: str,
) -> str:
    if measure_text_width(draw, font, f"{text}{suffix}") <= max_width:
        return text

    low = 0
    high = len(text)
    while low < high:
        middle = (low + high + 1) // 2
        if measure_text_width(draw, font, f"{text[:middle]}{suffix}") <= max_width:
            low = middle
        else:
            high = middle - 1

    return text[:low]


def wrap_text_to_width(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    text: str,
    max_width: int,
) -> list[str]:
    wrapped_lines: list[str] = []

    def append_fitted_segments(line: str) -> None:
        if measure_text_width(draw, font, line) <= max_width:
            wrapped_lines.append(line)
            return

        remainder = line
        while remainder:
            segment = fit_text_to_width(draw, font, remainder, max_width, suffix="")
            if not segment:
                wrapped_lines.append(line)
                return
            wrapped_lines.append(segment)
            remainder = remainder[len(segment) :]

    for paragraph in text.splitlines() or [""]:
        if not paragraph:
            wrapped_lines.append("")
            continue

        words = paragraph.split()
        if not words:
            wrapped_lines.append("")
            continue

        line = words[0]
        for word in words[1:]:
            candidate = f"{line} {word}"
            if measure_text_width(draw, font, candidate) <= max_width:
                line = candidate
            else:
                append_fitted_segments(line)
                line = word

        append_fitted_segments(line)

    return wrapped_lines


def quantise_to_kindle_greyscale(image: Image.Image) -> Image.Image:
    greyscale_image = image.convert("L")
    return greyscale_image.point(
        lambda value: max(
            0,
            min(
                255,
                round(round(value / KINDLE_GREYSCALE_STEP) * KINDLE_GREYSCALE_STEP),
            ),
        ),
    )


def scale_thin_shape_stroke(render_scale: int) -> int:
    return max(1, round(THIN_SHAPE_FINAL_STROKE_WIDTH * render_scale))


def finalise_dashboard_image(
    image: Image.Image,
    width: int,
    height: int,
) -> Image.Image:
    image = image.resize((width, height), resample=Image.Resampling.LANCZOS)
    return quantise_to_kindle_greyscale(image)



from datetime import UTC, datetime
from typing import TYPE_CHECKING

from PIL import Image, ImageDraw, ImageFont

from .chart_context import build_chart_context
from .cloud_cover import draw_cloud_cover_chart
from .common import (
    COLOUR_BLACK,
    COLOUR_WHITE,
    ChartBounds,
    build_chart_bounds,
    build_render_metrics,
    finalise_dashboard_image,
)
from .feedback import WEATHER_LOADING, draw_forecast_feedback
from .forecast import build_next_rain_header_message, normalise_chart_now
from .header import draw_header
from .precipitation import draw_precipitation_chart
from .temperature import draw_temperature_chart

if TYPE_CHECKING:
    from kindle_dashboard.weather import Forecast


def render_dashboard(
    width: int,
    height: int,
    *,
    battery_level: int | None,
    forecast: Forecast | None,
    forecast_error_message: str | None,
    weather_configured: bool,
    weather_data_stale: bool,
    supersample_scale: int,
    weather_data_stale_hours: int | None = None,
    next_rain_gap_tolerance_hours: float = 1.0,
) -> Image.Image:
    scaled_width = width * supersample_scale
    scaled_height = height * supersample_scale

    metrics = build_render_metrics(width, height, supersample_scale)

    image = Image.new("L", (scaled_width, scaled_height), color=COLOUR_WHITE)
    draw = ImageDraw.Draw(image)
    font = ImageFont.load_default(size=16 * supersample_scale)

    draw.rectangle(
        (0, 0, scaled_width - 1, scaled_height - 1),
        outline=COLOUR_BLACK,
        width=metrics.border_width,
    )

    now = datetime.now(tz=UTC)
    chart_now = normalise_chart_now(now, forecast) if forecast is not None else now
    weather_status_message = None
    if forecast is not None:
        weather_status_message = build_next_rain_header_message(
            forecast,
            chart_now,
            rain_gap_tolerance_hours=next_rain_gap_tolerance_hours,
        )
    elif weather_configured and forecast_error_message is None:
        weather_status_message = WEATHER_LOADING

    _, header_bottom = draw_header(
        draw=draw,
        font=font,
        scaled_width=scaled_width,
        now=now,
        battery_level=battery_level,
        weather_data_stale=weather_data_stale,
        weather_data_stale_hours=weather_data_stale_hours,
        weather_status_message=weather_status_message,
        metrics=metrics,
    )

    base_chart_bounds = build_chart_bounds(scaled_width, scaled_height, metrics)
    chart_bounds = ChartBounds(
        left=base_chart_bounds.left,
        top=max(base_chart_bounds.top, header_bottom + metrics.padding),
        right=base_chart_bounds.right,
        bottom=base_chart_bounds.bottom,
    )

    if forecast_error_message is not None:
        draw_forecast_feedback(
            draw=draw,
            font=font,
            left=chart_bounds.left,
            top=chart_bounds.top,
            right=chart_bounds.right,
            bottom=chart_bounds.bottom,
            message=forecast_error_message,
            error_text_line_spacing=metrics.error_text_line_spacing,
        )
    elif forecast is not None:
        chart_context = build_chart_context(
            draw=draw,
            font=font,
            chart_bounds=chart_bounds,
            now=chart_now,
            metrics=metrics,
        )
        draw_precipitation_chart(
            draw=draw,
            font=font,
            chart_bounds=chart_bounds,
            chart_context=chart_context,
            forecast=forecast,
            metrics=metrics,
        )
        draw_cloud_cover_chart(
            draw=draw,
            chart_bounds=chart_bounds,
            chart_context=chart_context,
            forecast=forecast,
        )
        draw_temperature_chart(
            draw=draw,
            font=font,
            chart_bounds=chart_bounds,
            chart_context=chart_context,
            forecast=forecast,
            metrics=metrics,
        )
    return finalise_dashboard_image(image, width, height)



from typing import TYPE_CHECKING

from .common import (
    COLOUR_BLACK,
    COLOUR_WHITE,
    KINDLE_GREYSCALE_LEVELS,
    ChartBounds,
    FontType,
    RenderMetrics,
    build_vertical_band,
    measure_text,
)
from .forecast import build_hourly_series

if TYPE_CHECKING:
    from PIL import ImageDraw

    from kindle_dashboard.weather import Forecast

    from .chart_context import ChartRenderContext

PRECIPITATION_PLOT_HEIGHT_RATIO = 0.30
PRECIPITATION_MAX = 2.0
PRECIPITATION_RULE_INTERVAL_MM = 0.2
PRECIPITATION_RULE_MAX_MM = 1.0
PRECIPITATION_RULE_COLOUR = 223
KINDLE_DARKEST_NON_BLACK_LEVEL = 1
KINDLE_LIGHTEST_NON_WHITE_LEVEL = KINDLE_GREYSCALE_LEVELS - 2


def clamp_precipitation(value: float) -> float:
    return max(0.0, min(PRECIPITATION_MAX, value))


def clamp_percentage(value: float) -> float:
    return max(0.0, min(100.0, value))


def probability_to_kindle_greyscale(probability: float) -> int:
    clamped_probability = clamp_percentage(probability)
    max_level = KINDLE_GREYSCALE_LEVELS - 1
    usable_range = KINDLE_LIGHTEST_NON_WHITE_LEVEL - KINDLE_DARKEST_NON_BLACK_LEVEL
    grey_level = KINDLE_LIGHTEST_NON_WHITE_LEVEL - round(
        (clamped_probability / 100.0) * usable_range
    )
    return round((grey_level / max_level) * COLOUR_WHITE)


def showers_dominate(rain: float, showers: float) -> bool:
    return showers > rain


def format_precipitation_label(value: float) -> str:
    if value > PRECIPITATION_MAX:
        return f">{PRECIPITATION_MAX:.1f}mm"
    return f"{clamp_precipitation(value):.1f}mm"


def choose_checker_cell_size(
    bar_width: int,
    preferred_cell_size: int,
) -> int:
    if bar_width <= 0:
        msg = "Bar width must be positive"
        raise ValueError(msg)

    preferred = max(1, min(preferred_cell_size, bar_width))
    candidate_sizes = [
        cell_size
        for cell_size in range(1, bar_width + 1)
        if bar_width % cell_size == 0
    ]
    return min(
        candidate_sizes,
        key=lambda cell_size: abs(cell_size - preferred),
    )


def draw_checkered_overlay(
    draw: ImageDraw.ImageDraw,
    *,
    left: int,
    top: int,
    right: int,
    bottom: int,
    cell_size: int,
) -> None:
    for y in range(top, bottom + 1, cell_size):
        row = (y - top) // cell_size
        for x in range(left, right + 1, cell_size):
            column = (x - left) // cell_size
            if (row + column) % 2 != 0:
                continue
            square_right = x + cell_size - 1
            square_bottom = min(bottom, y + cell_size - 1)
            draw.rectangle(
                (x, y, square_right, square_bottom),
                fill=COLOUR_WHITE,
                outline=COLOUR_WHITE,
            )


def draw_precipitation_chart(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    chart_bounds: ChartBounds,
    chart_context: ChartRenderContext,
    forecast: Forecast,
    metrics: RenderMetrics,
) -> None:
    hour_slots = chart_context.hour_slots
    x_axis_y = chart_context.x_axis_y
    x_positions = chart_context.x_positions

    # Open-Meteo's hourly precipitation value at HH:00 covers the preceding
    # interval [HH-1:00, HH:00), so bars are placed on the starting hour.
    slot_values = build_hourly_series(
        forecast,
        forecast.precipitation,
        hour_slots,
        hour_shift=-1,
    )
    slot_probabilities = build_hourly_series(
        forecast,
        forecast.precipitation_probability,
        hour_slots,
        hour_shift=-1,
    )
    slot_rain = build_hourly_series(
        forecast,
        forecast.rain,
        hour_slots,
        hour_shift=-1,
    )
    slot_showers = build_hourly_series(
        forecast,
        forecast.showers,
        hour_slots,
        hour_shift=-1,
    )
    precipitation_top, precipitation_bottom = build_vertical_band(
        chart_bounds.top,
        chart_bounds.bottom,
        top_ratio=1.0 - PRECIPITATION_PLOT_HEIGHT_RATIO,
        bottom_ratio=1.0,
    )
    plot_left = chart_bounds.left
    plot_top = precipitation_top + metrics.plot_padding
    plot_right = chart_bounds.right
    plot_bottom = min(
        precipitation_bottom - metrics.plot_padding,
        x_axis_y,
    )
    plot_height = plot_bottom - plot_top
    interval_values = slot_values[:-1]
    interval_probabilities = slot_probabilities[:-1]
    interval_rain = slot_rain[:-1]
    interval_showers = slot_showers[:-1]
    intervals = zip(
        interval_values,
        interval_probabilities,
        interval_rain,
        interval_showers,
        strict=False,
    )
    max_visible_precipitation = max(
        (clamp_precipitation(float(value)) for value in interval_values),
        default=0.0,
    )
    max_rule_value = min(max_visible_precipitation, PRECIPITATION_RULE_MAX_MM)
    if max_rule_value >= PRECIPITATION_RULE_INTERVAL_MM:
        rule_value = PRECIPITATION_RULE_INTERVAL_MM
        while rule_value <= max_rule_value + 1e-9:
            if abs(rule_value - PRECIPITATION_RULE_MAX_MM) < 1e-9:
                rule_value = round(rule_value + PRECIPITATION_RULE_INTERVAL_MM, 1)
                continue
            rule_y = plot_bottom - round(
                (rule_value / PRECIPITATION_MAX) * plot_height
            )
            if plot_top <= rule_y <= plot_bottom:
                draw.line(
                    (plot_left, rule_y, plot_right, rule_y),
                    fill=PRECIPITATION_RULE_COLOUR,
                    width=max(1, metrics.thin_shape_stroke_width),
                )
            rule_value = round(rule_value + PRECIPITATION_RULE_INTERVAL_MM, 1)

    largest_bar_label: tuple[str, int, int, int] | None = None
    largest_clamped_value = -1.0

    for index, (precipitation_value, probability_value, rain_value, showers_value) in (
        enumerate(intervals)
    ):
        value = float(precipitation_value)
        clamped_value = clamp_precipitation(value)
        bar_left = x_positions[index]
        bar_right = x_positions[index + 1]

        bar_height = round((clamped_value / PRECIPITATION_MAX) * plot_height)
        if clamped_value > 0 and bar_height == 0:
            bar_height = 1
        if bar_height <= 0:
            continue

        bar_top = plot_bottom - bar_height + 1
        bar_fill = probability_to_kindle_greyscale(float(probability_value))
        draw.rectangle(
            (bar_left, bar_top, bar_right, plot_bottom),
            fill=bar_fill,
            outline=bar_fill,
        )

        if clamped_value > largest_clamped_value:
            largest_clamped_value = clamped_value
            largest_bar_label = (
                format_precipitation_label(value),
                bar_left,
                bar_right,
                bar_top,
            )

        if showers_dominate(rain_value, showers_value):
            base_cell_size = max(2, metrics.thin_shape_stroke_width * 2)
            bar_width = bar_right - bar_left + 1
            checker_cell_size = choose_checker_cell_size(
                bar_width,
                max(2, base_cell_size + 1),
            )
            draw_checkered_overlay(
                draw,
                left=bar_left,
                top=bar_top,
                right=bar_right,
                bottom=plot_bottom,
                cell_size=checker_cell_size,
            )

    if max_rule_value >= PRECIPITATION_RULE_MAX_MM:
        major_rule_y = plot_bottom - round(
            (PRECIPITATION_RULE_MAX_MM / PRECIPITATION_MAX) * plot_height
        )
        if plot_top <= major_rule_y <= plot_bottom:
            draw.line(
                (plot_left, major_rule_y, plot_right, major_rule_y),
                fill=PRECIPITATION_RULE_COLOUR,
                width=max(1, metrics.thin_shape_stroke_width),
            )

    if largest_bar_label is None:
        return

    label_text, bar_left, bar_right, bar_top = largest_bar_label
    label_width, label_height, _ = measure_text(draw, font, label_text)
    label_x = bar_left + ((bar_right - bar_left + 1 - label_width) // 2)
    label_x = max(chart_bounds.left, min(label_x, chart_bounds.right - label_width))
    label_y = max(
        chart_bounds.top,
        bar_top - label_height - metrics.temperature_label_offset,
    )
    draw.text((label_x, label_y), label_text, fill=COLOUR_BLACK, font=font)



import os
from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class ServerConfig:
    host: str
    port: int
    default_width: int
    default_height: int
    render_supersample_scale: int
    weather_update_secs: int
    next_rain_gap_tolerance_hours: float
    weather_latitude: float | None
    weather_longitude: float | None


def load_server_config() -> ServerConfig:
    weather_latitude_raw = os.getenv("WEATHER_LATITUDE")
    weather_longitude_raw = os.getenv("WEATHER_LONGITUDE")
    if (weather_latitude_raw is None) != (weather_longitude_raw is None):
        message = "WEATHER_LATITUDE and WEATHER_LONGITUDE must both be set together"
        raise ValueError(message)
    next_rain_gap_tolerance_hours = float(
        os.getenv("NEXT_RAIN_GAP_TOLERANCE_HOURS", "1")
    )
    if next_rain_gap_tolerance_hours < 0:
        message = "NEXT_RAIN_GAP_TOLERANCE_HOURS must be non-negative"
        raise ValueError(message)
    return ServerConfig(
        host=os.getenv("HOST", "127.0.0.1"),
        port=int(os.getenv("PORT", "42137")),
        default_width=int(os.getenv("IMAGE_WIDTH", "600")),
        default_height=int(os.getenv("IMAGE_HEIGHT", "800")),
        render_supersample_scale=int(os.getenv("RENDER_SUPERSAMPLE_SCALE", "2")),
        weather_update_secs=int(os.getenv("WEATHER_UPDATE_SECS", str(30 * 60))),
        next_rain_gap_tolerance_hours=next_rain_gap_tolerance_hours,
        weather_latitude=float(weather_latitude_raw)
        if weather_latitude_raw is not None
        else None,
        weather_longitude=float(weather_longitude_raw)
        if weather_longitude_raw is not None
        else None,
    )



import logging
import os
import sys

_SYSLOG_PRIORITIES = {
    logging.CRITICAL: 2,
    logging.ERROR: 3,
    logging.WARNING: 4,
    logging.INFO: 6,
    logging.DEBUG: 7,
}


def _parse_log_level(raw_level: str) -> int:
    level_name = raw_level.strip().upper()
    if not level_name:
        message = "LOG_LEVEL must not be empty"
        raise ValueError(message)

    level = logging.getLevelNamesMapping().get(level_name)
    if level is None:
        message = f"LOG_LEVEL has unknown level: {raw_level}"
        raise ValueError(message)
    return level


def _read_bool_env(name: str) -> bool | None:
    value = os.getenv(name)
    if value is None:
        return None

    normalized = value.strip().lower()
    if normalized in {"1", "true", "yes", "on"}:
        return True
    if normalized in {"0", "false", "no", "off"}:
        return False
    message = f"{name} must be a boolean value"
    raise ValueError(message)


def _use_systemd_logging() -> bool:
    force_systemd = _read_bool_env("LOG_SYSTEMD")
    if force_systemd is not None:
        return force_systemd
    return os.getenv("JOURNAL_STREAM") is not None


class _SystemdFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        message = super().format(record)
        priority = _SYSLOG_PRIORITIES.get(
            record.levelno,
            7 if record.levelno < logging.INFO else 6,
        )
        return f"<{priority}>{message}"


def configure_logging() -> None:
    log_level = _parse_log_level(os.getenv("LOG_LEVEL", "INFO"))
    handler = logging.StreamHandler(stream=sys.stdout)

    formatter = (
        _SystemdFormatter("%(name)s: %(message)s")
        if _use_systemd_logging()
        else logging.Formatter(
            "%(asctime)s %(levelname)s %(name)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )

    handler.setFormatter(formatter)

    root_logger = logging.getLogger()
    root_logger.handlers.clear()
    root_logger.setLevel(log_level)
    root_logger.addHandler(handler)



from typing import TYPE_CHECKING

from .common import (
    COLOUR_BLACK,
    COLOUR_WHITE,
    TEXT_ELLIPSIS,
    FontType,
    RenderMetrics,
    fit_text_to_width,
    measure_text,
)

if TYPE_CHECKING:
    from datetime import datetime

    from PIL import ImageDraw

BATTERY_TERMINAL_DIMENSION_MIN = 2
BATTERY_TERMINAL_WIDTH_DIVISOR = 12
BATTERY_TERMINAL_HEIGHT_DIVISOR = 2
BATTERY_INNER_PADDING_MIN = 2


def build_weather_stale_message(stale_hours: int | None) -> str:
    display_hours = stale_hours if stale_hours is not None else 1
    units = "hour" if display_hours == 1 else "hours"
    return f"Warning: weather data is {display_hours} {units} stale"


def draw_battery_icon(
    draw: ImageDraw.ImageDraw,
    x: int,
    y: int,
    width: int,
    height: int,
    outline_width: int,
    battery_level: int,
) -> None:
    terminal_width = max(
        BATTERY_TERMINAL_DIMENSION_MIN,
        width // BATTERY_TERMINAL_WIDTH_DIVISOR,
    )
    body_right = x + width - terminal_width - 1
    body_bottom = y + height - 1

    draw.rectangle(
        (x, y, body_right, body_bottom),
        outline=COLOUR_BLACK,
        width=outline_width,
        fill=COLOUR_WHITE,
    )

    terminal_height = max(
        BATTERY_TERMINAL_DIMENSION_MIN,
        height // BATTERY_TERMINAL_HEIGHT_DIVISOR,
    )
    terminal_top = y + (height - terminal_height) // 2
    draw.rectangle(
        (
            body_right + 1,
            terminal_top,
            x + width - 1,
            terminal_top + terminal_height - 1,
        ),
        outline=COLOUR_BLACK,
        fill=COLOUR_BLACK,
    )

    inner_padding = max(BATTERY_INNER_PADDING_MIN, outline_width + 1)
    inner_left = x + inner_padding
    inner_top = y + inner_padding
    inner_right = body_right - inner_padding
    inner_bottom = body_bottom - inner_padding
    if inner_left > inner_right or inner_top > inner_bottom:
        return

    inner_width = inner_right - inner_left + 1
    fill_width = round(inner_width * (battery_level / 100))
    if fill_width <= 0:
        return

    draw.rectangle(
        (inner_left, inner_top, inner_left + fill_width - 1, inner_bottom),
        fill=COLOUR_BLACK,
    )


def draw_header(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    scaled_width: int,
    now: datetime,
    *,
    battery_level: int | None,
    weather_data_stale: bool,
    weather_data_stale_hours: int | None,
    weather_status_message: str | None,
    metrics: RenderMetrics,
) -> tuple[int, int]:
    top_y = metrics.border_width + metrics.padding
    left_edge = metrics.border_width + metrics.padding
    right_edge = scaled_width - left_edge

    clock_label = now.strftime("%H:%M")
    clock_text_width, clock_text_height, _ = measure_text(draw, font, clock_label)
    clock_x = max(left_edge, right_edge - clock_text_width)
    header_bottom = top_y + clock_text_height

    if battery_level is not None:
        battery_label = f"{battery_level}%"
        (
            battery_text_width,
            battery_text_height,
            battery_text_top_offset,
        ) = measure_text(draw, font, battery_label)
        battery_icon_height = max(
            metrics.battery_icon_min_height,
            battery_text_height + metrics.battery_icon_text_padding,
        )
        battery_icon_width = battery_icon_height * 2
        right_group_width = (
            battery_text_width
            + metrics.group_spacing
            + battery_icon_width
            + metrics.group_spacing
            + clock_text_width
        )
        battery_text_x = max(left_edge, right_edge - right_group_width)
        battery_icon_x = battery_text_x + battery_text_width + metrics.group_spacing
        battery_icon_y = (
            top_y
            + battery_text_top_offset
            + (battery_text_height - battery_icon_height) // 2
        )
        clock_x = battery_icon_x + battery_icon_width + metrics.group_spacing

        draw.text((battery_text_x, top_y), battery_label, fill=COLOUR_BLACK, font=font)
        draw_battery_icon(
            draw=draw,
            x=battery_icon_x,
            y=battery_icon_y,
            width=battery_icon_width,
            height=battery_icon_height,
            outline_width=metrics.thin_shape_stroke_width,
            battery_level=battery_level,
        )
        header_bottom = max(header_bottom, battery_icon_y + battery_icon_height)

    draw.text((clock_x, top_y), clock_label, fill=COLOUR_BLACK, font=font)

    header_left_message: str | None = None
    if weather_data_stale:
        header_left_message = build_weather_stale_message(weather_data_stale_hours)
    elif weather_status_message:
        header_left_message = weather_status_message

    if header_left_message:
        warning_max_width = max(0, clock_x - left_edge - metrics.group_spacing)
        warning_text = fit_text_to_width(
            draw,
            font,
            header_left_message,
            warning_max_width,
            suffix=TEXT_ELLIPSIS,
        )
        if warning_text:
            if warning_text != header_left_message:
                warning_text = f"{warning_text}{TEXT_ELLIPSIS}"
            _, warning_height, _ = measure_text(draw, font, warning_text)
            draw.text((left_edge, top_y), warning_text, fill=COLOUR_BLACK, font=font)
            header_bottom = max(header_bottom, top_y + warning_height)

    return top_y, header_bottom



from bisect import bisect_left
from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from kindle_dashboard.weather import Forecast

RAIN_THRESHOLD_MM = 0.2
RAIN_EVENT_GAP_TOLERANCE_HOURS = 1.0
NO_RAIN_LABEL = "No rain in 7d"


def build_next_rain_header_message(
    forecast: Forecast,
    now: datetime,
    *,
    rain_threshold_mm: float = RAIN_THRESHOLD_MM,
    rain_gap_tolerance_hours: float = RAIN_EVENT_GAP_TOLERANCE_HOURS,
) -> str | None:
    next_window = build_next_hourly_rain_window(
        forecast,
        now,
        rain_threshold_mm=rain_threshold_mm,
        rain_gap_tolerance_hours=rain_gap_tolerance_hours,
    )
    if next_window is not None:
        start_time, end_time = next_window
        raining_now = start_time <= now < end_time
        start_label = "now" if raining_now else start_time.strftime("%H:%M")
        last_timestamp = (
            forecast.hourly_datetimes[-1] if forecast.hourly_datetimes else None
        )
        if last_timestamp is not None and end_time >= last_timestamp:
            return (
                "Raining all day"
                if raining_now
                else f"Next Rain {start_label} onwards"
            )
        if raining_now:
            return f"Raining until {end_time.strftime('%H:%M')}"
        return f"Next Rain {start_label} to {end_time.strftime('%H:%M')}"

    next_day = build_next_rain_day(forecast, now, rain_threshold_mm=rain_threshold_mm)
    if next_day is None:
        return NO_RAIN_LABEL

    today = now.date()
    if next_day == today:
        return "Next Rain today"
    if next_day == (today + timedelta(days=1)):
        return "Next Rain tomorrow"
    return f"Next Rain {next_day.strftime('%a')}"


def build_next_hourly_rain_window(
    forecast: Forecast,
    now: datetime,
    *,
    rain_threshold_mm: float,
    rain_gap_tolerance_hours: float = RAIN_EVENT_GAP_TOLERANCE_HOURS,
) -> tuple[datetime, datetime] | None:
    window_start: datetime | None = None
    window_end: datetime | None = None
    gap_tolerance = timedelta(hours=rain_gap_tolerance_hours)

    for interval_end, precipitation_value in zip(
        forecast.hourly_datetimes,
        forecast.precipitation,
        strict=False,
    ):
        if precipitation_value < rain_threshold_mm:
            continue

        if interval_end <= now:
            continue

        interval_start = interval_end - timedelta(hours=1)
        if window_start is None:
            window_start = interval_start
            window_end = interval_end
            continue
        assert window_end is not None
        if interval_start > (window_end + gap_tolerance):
            break
        window_end = max(window_end, interval_end)

    if window_start is None or window_end is None:
        return None

    return window_start, window_end


def build_next_rain_day(
    forecast: Forecast,
    now: datetime,
    *,
    rain_threshold_mm: float,
) -> date | None:
    for day, precipitation_sum in zip(
        forecast.daily_dates,
        forecast.daily_precipitation_sum,
        strict=False,
    ):
        if precipitation_sum < rain_threshold_mm:
            continue
        if day >= now.date():
            return day

    return None


def build_sun_event_markers(
    forecast: Forecast,
    window_start: datetime,
    window_end: datetime,
) -> list[tuple[datetime, bool]]:
    markers = [
        (sunrise, True)
        for sunrise in forecast.daily_sunrise_datetimes
        if window_start <= sunrise <= window_end
    ]
    markers.extend(
        (sunset, False)
        for sunset in forecast.daily_sunset_datetimes
        if window_start <= sunset <= window_end
    )

    markers.sort(key=lambda marker: marker[0])
    return markers


def build_hourly_series(
    forecast: Forecast,
    values: list[float],
    hour_slots: list[datetime],
    *,
    hour_shift: int,
) -> list[float]:
    lookup: dict[datetime, float] = {}
    shift_delta = timedelta(hours=hour_shift)

    for timestamp, value in zip(forecast.hourly_datetimes, values, strict=False):
        shifted_timestamp = timestamp + shift_delta
        key = shifted_timestamp.replace(minute=0, second=0, microsecond=0)
        lookup[key] = value

    assert lookup, "Forecast series did not provide any hourly values"

    missing = object()
    known_slots: list[datetime] | None = None
    known_values: list[float] | None = None
    series: list[float] = []

    for slot in hour_slots:
        value = lookup.get(slot, missing)
        if value is not missing:
            series.append(cast("float", value))
            continue

        if known_slots is None:
            known_slots = sorted(lookup)
            known_values = [lookup[known_slot] for known_slot in known_slots]
        assert known_values is not None

        series.append(_nearest_slot_value(slot, known_slots, known_values))

    return series


def _nearest_slot_value(
    slot: datetime,
    known_slots: list[datetime],
    known_values: list[float],
) -> float:
    idx = bisect_left(known_slots, slot)

    if idx == 0:
        return known_values[0]
    if idx == len(known_slots):
        return known_values[-1]

    previous_slot = known_slots[idx - 1]
    next_slot = known_slots[idx]
    previous_gap = slot - previous_slot
    next_gap = next_slot - slot

    # Keep deterministic behaviour for exact ties by preferring the earlier slot.
    if previous_gap <= next_gap:
        return known_values[idx - 1]
    return known_values[idx]


def normalise_chart_now(now: datetime, forecast: Forecast) -> datetime:
    return now.astimezone(forecast.forecast_timezone).replace(tzinfo=None)



"""Retrieve hourly weather data from Open-Meteo."""

import json
import logging
from dataclasses import dataclass, field
from datetime import date, datetime
from time import perf_counter
from typing import cast
from urllib.error import URLError
from urllib.parse import urlencode
from urllib.request import urlopen
from zoneinfo import ZoneInfo

OPEN_METEO_URL = "https://api.open-meteo.com/v1/forecast"
DEFAULT_TIMEOUT_SECONDS = 10
logger = logging.getLogger(__name__)


class WeatherFetchError(RuntimeError):
    """Raised when weather data cannot be fetched from the remote API."""


@dataclass(frozen=True, slots=True)
class Forecast:
    time: list[str]
    temperature_2m: list[float]
    precipitation: list[float]
    precipitation_probability: list[float]
    rain: list[float]
    showers: list[float]
    cloudcover_low: list[float]
    cloudcover_mid: list[float]
    timezone: str
    utc_offset_seconds: int | None
    daily_time: list[str] = field(default_factory=list)
    daily_precipitation_sum: list[float] = field(default_factory=list)
    daily_sunrise: list[str] = field(default_factory=list)
    daily_sunset: list[str] = field(default_factory=list)
    hourly_datetimes: tuple[datetime, ...] = field(init=False, repr=False)
    daily_dates: tuple[date, ...] = field(init=False, repr=False)
    daily_sunrise_datetimes: tuple[datetime, ...] = field(init=False, repr=False)
    daily_sunset_datetimes: tuple[datetime, ...] = field(init=False, repr=False)
    forecast_timezone: ZoneInfo = field(init=False, repr=False)

    def __post_init__(self) -> None:
        forecast_timezone = _resolve_forecast_timezone(self.timezone)
        object.__setattr__(self, "forecast_timezone", forecast_timezone)
        object.__setattr__(
            self,
            "hourly_datetimes",
            _parse_forecast_datetimes(self.time, forecast_timezone),
        )
        object.__setattr__(
            self,
            "daily_dates",
            _parse_forecast_dates(self.daily_time),
        )
        object.__setattr__(
            self,
            "daily_sunrise_datetimes",
            _parse_forecast_datetimes(self.daily_sunrise, forecast_timezone),
        )
        object.__setattr__(
            self,
            "daily_sunset_datetimes",
            _parse_forecast_datetimes(self.daily_sunset, forecast_timezone),
        )


def _resolve_forecast_timezone(
    timezone_name: str,
) -> ZoneInfo:
    return ZoneInfo(timezone_name)


def _parse_forecast_datetime(
    timestamp_text: str,
    forecast_timezone: ZoneInfo,
) -> datetime:
    timestamp = datetime.fromisoformat(timestamp_text)

    if timestamp.tzinfo is not None:
        timestamp = timestamp.astimezone(forecast_timezone).replace(tzinfo=None)

    return timestamp


def _parse_forecast_date(timestamp_text: str) -> date:
    return date.fromisoformat(timestamp_text)


def _parse_forecast_datetimes(
    timestamps: list[str],
    forecast_timezone: ZoneInfo,
) -> tuple[datetime, ...]:
    return tuple(
        _parse_forecast_datetime(timestamp_text, forecast_timezone)
        for timestamp_text in timestamps
    )


def _parse_forecast_dates(timestamps: list[str]) -> tuple[date, ...]:
    return tuple(_parse_forecast_date(timestamp_text) for timestamp_text in timestamps)


def _coerce_float_series(values: object) -> list[float]:
    return [float(value) for value in cast("list[float | int | str]", values)]


def _extract_float_series(
    hourly: dict[str, object], *candidate_keys: str
) -> list[float]:
    for key in candidate_keys:
        values = hourly.get(key)
        if values is None:
            continue
        return _coerce_float_series(values)

    missing_keys = ", ".join(candidate_keys)
    msg = f"Missing expected hourly series: {missing_keys}"
    raise KeyError(msg)


def fetch_forecast_data(latitude: float, longitude: float) -> Forecast:
    query = urlencode(
        {
            "latitude": latitude,
            "longitude": longitude,
            "hourly": (
                "temperature_2m,precipitation,precipitation_probability,"
                "rain,showers,cloud_cover_low,cloud_cover_mid"
            ),
            "daily": "precipitation_sum,sunrise,sunset",
            "timezone": "auto",
            "forecast_hours": 24,
            "forecast_days": 7,
        }
    )
    url = f"{OPEN_METEO_URL}?{query}"
    start_time = perf_counter()

    try:
        with urlopen(url, timeout=DEFAULT_TIMEOUT_SECONDS) as response:  # noqa: S310
            payload = response.read()
    except URLError as exc:
        elapsed_ms = (perf_counter() - start_time) * 1000
        logger.warning(
            "Weather request failed after %.1f ms (latitude=%s, longitude=%s): %s",
            elapsed_ms,
            latitude,
            longitude,
            exc,
        )
        message = f"Unable to fetch weather:\n{exc}"
        raise WeatherFetchError(message) from exc

    elapsed_ms = (perf_counter() - start_time) * 1000
    logger.info(
        "Weather request completed in %.1f ms (latitude=%s, longitude=%s)",
        elapsed_ms,
        latitude,
        longitude,
    )

    # We deliberately trust Open-Meteo's payload contract and do not
    # validate schema here.
    raw_data = cast("dict[str, object]", json.loads(payload))
    hourly = cast("dict[str, object]", raw_data["hourly"])
    daily = cast("dict[str, object]", raw_data["daily"])
    timezone_name = cast("str", raw_data["timezone"])
    utc_offset_seconds = cast("int | float | None", raw_data["utc_offset_seconds"])

    return Forecast(
        time=cast("list[str]", hourly["time"]),
        temperature_2m=_coerce_float_series(hourly["temperature_2m"]),
        precipitation=_coerce_float_series(hourly["precipitation"]),
        precipitation_probability=_coerce_float_series(
            hourly["precipitation_probability"]
        ),
        rain=_coerce_float_series(hourly["rain"]),
        showers=_coerce_float_series(hourly["showers"]),
        cloudcover_low=_extract_float_series(
            hourly,
            "cloud_cover_low",
            "cloudcover_low",
        ),
        cloudcover_mid=_extract_float_series(
            hourly,
            "cloud_cover_mid",
            "cloudcover_mid",
        ),
        timezone=timezone_name,
        utc_offset_seconds=(
            int(utc_offset_seconds) if utc_offset_seconds is not None else None
        ),
        daily_time=cast("list[str]", daily["time"]),
        daily_precipitation_sum=_coerce_float_series(daily["precipitation_sum"]),
        daily_sunrise=cast("list[str]", daily["sunrise"]),
        daily_sunset=cast("list[str]", daily["sunset"]),
    )



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from .common import (
    ChartBounds,
    FontType,
    RenderMetrics,
    build_x_positions,
    measure_text,
)

if TYPE_CHECKING:
    from PIL import ImageDraw

TEMPERATURE_LABEL_SAMPLE = "00"
CHART_HOURS = 24


@dataclass(frozen=True)
class ChartRenderContext:
    hour_slots: list[datetime]
    x_axis_y: int
    x_positions: list[int]


def compute_x_axis_y(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    chart_bottom: int,
    plot_padding: int,
) -> int:
    _, label_height, _ = measure_text(draw, font, TEMPERATURE_LABEL_SAMPLE)
    return round(chart_bottom - label_height - (3 * plot_padding))


def build_chart_context(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    chart_bounds: ChartBounds,
    now: datetime,
    metrics: RenderMetrics,
) -> ChartRenderContext:
    start_hour = now.replace(minute=0, second=0, microsecond=0)
    hour_slots = [
        start_hour + timedelta(hours=offset) for offset in range(CHART_HOURS)
    ]
    x_axis_y = compute_x_axis_y(
        draw=draw,
        font=font,
        chart_bottom=chart_bounds.bottom,
        plot_padding=metrics.plot_padding,
    )
    x_positions = build_x_positions(
        chart_bounds.left,
        chart_bounds.right,
        len(hour_slots),
    )
    return ChartRenderContext(
        hour_slots=hour_slots,
        x_axis_y=x_axis_y,
        x_positions=x_positions,
    )



from typing import TYPE_CHECKING

from .common import (
    COLOUR_BLACK,
    TEXT_ELLIPSIS,
    FontType,
    fit_text_to_width,
    measure_text,
    wrap_text_to_width,
)

if TYPE_CHECKING:
    from PIL import ImageDraw

TEXT_LINE_HEIGHT_SAMPLE = "Ag"
FORECAST_UNAVAILABLE = "Weather unavailable"
WEATHER_LOADING = "Weather loading..."
UNKNOWN_WEATHER_ERROR = "Unknown weather error"


def draw_forecast_feedback(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    left: int,
    top: int,
    right: int,
    bottom: int,
    message: str,
    error_text_line_spacing: int,
) -> None:
    feedback_message = message.strip() or UNKNOWN_WEATHER_ERROR
    lines = [
        *wrap_text_to_width(
            draw,
            font,
            FORECAST_UNAVAILABLE,
            right - left + 1,
        ),
        "",
        *wrap_text_to_width(draw, font, feedback_message, right - left + 1),
    ]
    draw_wrapped_text_block(
        draw=draw,
        font=font,
        left=left,
        top=top,
        right=right,
        bottom=bottom,
        lines=lines,
        line_spacing=error_text_line_spacing,
    )


def draw_wrapped_text_block(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    left: int,
    top: int,
    right: int,
    bottom: int,
    lines: list[str],
    line_spacing: int,
) -> None:
    available_width = right - left + 1
    available_height = bottom - top + 1
    if available_width <= 0 or available_height <= 0:
        return

    _, line_height, _ = measure_text(draw, font, TEXT_LINE_HEIGHT_SAMPLE)
    if line_height <= 0:
        return

    max_lines = max(1, available_height // (line_height + line_spacing))
    if len(lines) > max_lines:
        lines = lines[:max_lines]
        trimmed = lines[-1].rstrip()
        fitted = fit_text_to_width(
            draw,
            font,
            trimmed,
            available_width,
            suffix=TEXT_ELLIPSIS,
        )
        lines[-1] = f"{fitted}{TEXT_ELLIPSIS}" if fitted else TEXT_ELLIPSIS

    y_position = top
    for line in lines:
        draw.text((left, y_position), line, fill=COLOUR_BLACK, font=font)
        y_position += line_height + line_spacing


def draw_forecast_loading(
    draw: ImageDraw.ImageDraw,
    font: FontType,
    left: int,
    top: int,
    right: int,
    bottom: int,
    error_text_line_spacing: int,
) -> None:
    lines = wrap_text_to_width(draw, font, WEATHER_LOADING, right - left + 1)
    draw_wrapped_text_block(
        draw=draw,
        font=font,
        left=left,
        top=top,
        right=right,
        bottom=bottom,
        lines=lines,
        line_spacing=error_text_line_spacing,
    )
